import processing.core.*; 
import processing.data.*; 
import processing.event.*; 
import processing.opengl.*; 

import java.util.HashMap; 
import java.util.ArrayList; 
import java.io.File; 
import java.io.BufferedReader; 
import java.io.PrintWriter; 
import java.io.InputStream; 
import java.io.OutputStream; 
import java.io.IOException; 

public class PixelSurgeon extends PApplet {

int screenX = 800;
int screenY = 800;

int scrollOffset = 128;

// How likely for a cell to be alive at start (in percentage)
float probabilityOfAliveAtStart = 15;

// Variables for timer
int interval = 50;
int lastRecordedTime = 0;

int waves = 0;

// Menu vars
// boolean playing = false;
boolean pause = true;
boolean gameWin = false;
boolean gameOver = false;

Input input;
Player player;
Level level;
AutomataCellular automataCellular;
Hud hud;

int cellSize = 4;
int newCellSize = 4;

public void setup() {
  size (screenX, screenY);
  noSmooth();

  Init();
  Clean();
}

public void Clean() {
  noStroke();
  background(0);
}

public void draw() {

  Clean();

  if (!pause) {
    UpdateGame();
  }

  UpdateMenu();

  if (newCellSize != cellSize) {
    cellSize = newCellSize;
    StartGame();
    pause = false;
  }
}

public void UpdateMenu() {
  hud.Update();
}

public void StartGame() {
  Init();
}

public void Init() {
  pause = true;
  gameOver = false;
  gameWin = false;
  ResetCount();
  waves = 0;

  input = new Input();
  player = new Player();
  level = new Level(); 
  automataCellular = new AutomataCellular();
  hud = new Hud();
}

public void UpdateGame() {

  // Iterate if timer ticks
  if (millis()-lastRecordedTime>interval) {
    automataCellular.iteration(level.cells);
    lastRecordedTime = millis();
    level.DrawLevel();
  }

  level.Update();

  player.Update();

  CheckGameOver();

  if (mousePressed && (mouseButton == RIGHT)) {
    level.FireLazer();
  }

}

public void CheckGameOver() {
  // println("ratio : "+cellCountVital / float(cellTotalVital) );
  //  ||
    //GetBadCellRatio() > 0.5
  if (GetVitalCellRatio() < 0.1f) {
    gameOver = true;
    pause = true;
  }
}

public void keyReleased() {

  input.KeyReleased(key, keyCode);

  // DEBUG
  if (key == 'o') level.SpawnBad();
  // if (key == 'o') gameOver = true;
  if (key == 'r') StartGame();
  if (key == '+') newCellSize = 8;
  if (key == '-') newCellSize = 4;
}

public void keyPressed() {

  input.KeyPressed(key, keyCode);

}

public void mouseReleased() {
  
  level.CleanLazer();

}
class AutomataCellular {

	int[][] cellsBuffer;

	AutomataCellular() {
		cellsBuffer = new int[level.imageData.width/cellSize][level.imageData.height/cellSize];
	}


	public void iteration(int[][] cells) { // When the clock ticks
  	// Save cells to buffer (so we opeate with one array keeping the other intact)
	  for (int x=0; x<level.imageData.width/cellSize; x++) {
	    for (int y=0; y<level.imageData.height/cellSize; y++) {
	      cellsBuffer[x][y] = cells[x][y];
	    }
	  }

	  // Visit each cell:
	  for (int x=0; x<level.imageData.width/cellSize; x++) {
	    for (int y=0; y<level.imageData.height/cellSize; y++) {

	      // And visit all the neighbours of each cell
	      int neighbours = 0; // We'll count the neighbours
	      for (int xx=x-1; xx<=x+1;xx++) {
	        for (int yy=y-1; yy<=y+1;yy++) {  
	          if (((xx>=0)&&(xx<level.imageData.width/cellSize))&&((yy>=0)&&(yy<level.imageData.height/cellSize))) { // Make sure you are not out of bounds
	            if (!((xx==x)&&(yy==y))) { // Make sure to to check against self
	              if (cellsBuffer[xx][yy] == cellStatusBad){
	                neighbours ++; // Check alive neighbours and count them
	              }
	            } // End of if
	          } // End of if
	        } // End of yy loop
	      } //End of xx loop
	      // We've checked the neigbours: apply rules!
	      if (cellsBuffer[x][y] == cellStatusBad) { // The cell is alive: kill it if necessary
	        if (neighbours < 2 || neighbours > 3) {
	          cells[x][y] = cellStatusDead; // Die unless it has 2 or 3 neighbours
	        }
	      } 
	      else  { // The cell is dead: make it live if necessary      
	        if (neighbours == 3 ) {
	          cells[x][y] = cellStatusBad; // Only if it has 3 neighbours
	        }
	      } // End of if
	    } // End of y loop
	  } // End of x loop
	} // End of function
}
class Bullet {

	int x = 0;
	int y = 0;
	int previousCellStatus = 0;

	boolean alive = true;

	PVector position;
	PVector direction;
	int speed = cellSize;

	float timeStarted = 0;
	float timeDelay = 2000;

	ArrayList<Cell> ghosts;

	final int MAX_GHOSTS = 3;

	Bullet(float x_, float y_, float dirX, float dirY, int previousCellStatus_) {
		Start(x_, y_, dirX, dirY, previousCellStatus_);
	}
	public void Start(float x_, float y_, float dirX, float dirY, int previousCellStatus_) {
		alive = true;
		position = new PVector(x_, y_);
		direction = new PVector(dirX, dirY);
		direction.normalize();
		// ghosts = new ArrayList<Cell>();
		timeStarted = millis();

		x = PApplet.parseInt(x_ / cellSize);
		y = PApplet.parseInt(y_ / cellSize);
		previousCellStatus = previousCellStatus_;
	}

	public void Kill() { 
		alive = false;
	}
	public boolean IsAlive() { return alive; }

}
class Cell {
	
	int status = 0;
	int x = 0;
	int y = 0;
	int cellColor;

	Cell (int x_, int y_, int cellStatus) {
		status = cellStatus;
		x = x_;
		y = y_;
	}
}
class Hud {

	PFont fontTitle, fontButton, fontParagraph;

	String explanation = "You are inside a human body ravaged by sickness.\nDestroy bad cells before they kill vital cells.";
	String controls = "Controls :\nMove with arrows\nAim with mouse\nShoot with clic\nHold mouse button for autofire";
	String credits = "Concept & Code : Leon Denise\nPixel Anatomy : Simon Denise";
	String references = "Inspired by Rick Levine's Microsurgeon and Conway's Game of Life";

	PVector anchor;

	final float JAUGE_HEIGHT = 8;
	final float JAUGE_MAX_WIDTH = 256;

	Hud() {

	  // Create the font
	  // printArray(PFont.list());
	  fontTitle = createFont("Cachiyuyo", 64, false);
	  fontButton = createFont("Cachiyuyo", 32, false);
	  fontParagraph = createFont("Cachiyuyo", 16, false);

	  anchor = new PVector(24, 24);
	}

	public void Update() {

		// Start Menu
		if (pause) {

			background(255);

			if (!gameOver) {

				textAlign(CENTER);
				fill(0);
				textFont(fontParagraph);
				text(explanation, screenX * 0.5f, screenY * 0.35f);
				text(controls, screenX * 0.5f, screenY * 0.6f);
				fill(100);
				text(credits, screenX * 0.5f, screenY * 0.85f);
				text(references, screenX * 0.5f, screenY * 0.95f);
				textFont(fontTitle);
				fill(0);
				text("PIXEL SURGEON", screenX * 0.5f, screenY * 0.25f);
				textFont(fontButton);
				fill(200, 0, 0);
				text("START", screenX * 0.5f, screenY * 0.5f);

		  		if (mousePressed && TestCollision(mouseX, mouseY, PApplet.parseInt(screenX * 0.5f - 80), PApplet.parseInt(screenY * 0.5f - 64), 160, 64)) {
		  			StartGame();
		  			pause = false;
		  		}

			} else {

				textAlign(CENTER);
				textFont(fontTitle);
				fill(0);
				text("Game Over", screenX * 0.5f, screenY * 0.5f);
				textFont(fontButton);
				fill(200, 0, 0);
				text("RESTART", screenX * 0.5f, screenY * 0.6f);

		  		if (mousePressed && TestCollision(mouseX, mouseY, PApplet.parseInt(screenX * 0.5f - 96), PApplet.parseInt(screenY * 0.6f - 64), 192, 64)) {
		  			StartGame();
		  			pause = false;
		  			gameOver = false;
		  		}
			}

	  	// In Game
	  	} else {

	  		// float[] ratios = GetRatioCells(); 
	  		// for (int c = 0; c < ratios.length; c++){
	  		// 	fill(GetColorByType(cellStatusHeart));
	  		// 	rect(anchor.x, anchor.y + c * (JAUGE_HEIGHT + JAUGE_HEIGHT * 0.5), ratios[c] * JAUGE_MAX_WIDTH, JAUGE_HEIGHT);
	  		// }
	  		// Waves
			textAlign(LEFT);
			fill(255);
			textFont(fontParagraph);
			text("left clic : heal\nright clic : lazer", anchor.x, anchor.y + JAUGE_HEIGHT);


	  		// Waves
			textAlign(RIGHT);
			fill(255);
			textFont(fontParagraph);
			text("waves : " + waves + "\nbad cells : " + cellCountBad/(cellSize*cellSize), screenX - anchor.x, anchor.y + JAUGE_HEIGHT);
	  	}
	}
}

IntDict keys;

class Input {


	Input() {
		keys = new IntDict();
		keys.set("left", 0);
		keys.set("right", 0);
		keys.set("up", 0);
		keys.set("down", 0);
		keys.set("space", 0);
	}

	public void KeyPressed(char key, int keyCode) {
		switch (keyCode){
			case UP:
				keys.set("up", 1);
				break;
			case DOWN:
				keys.set("down", 1);
				break;
			case LEFT:
				keys.set("left", 1);
				break;
			case RIGHT:
				keys.set("right", 1);
				break;
		}
		switch (key){
			case ' ':
				keys.set("space", 1);
				break;
		}
	}

	public void KeyReleased(char key, int keyCode) {
		switch (keyCode){
			case UP:
				keys.set("up", 0);
				break;
			case DOWN:
				keys.set("down", 0);
				break;
			case LEFT:
				keys.set("left", 0);
				break;
			case RIGHT:
				keys.set("right", 0);
				break;
		}
		switch (key){
			case ' ':
				keys.set("space", 0);
				break;
		}
	}
}
class Lazer {

	PVector direction;

	Lazer() {
		direction = new PVector();
	}

	public ArrayList<Cell> FireCells(int fromX, int fromY, int toX, int toY) {
		ArrayList<Cell> cells = new ArrayList<Cell>();
		float t = 0.0f;
		while(t < 1.0f) {
			int x = PApplet.parseInt(fromX + (toX - fromX) * t);
			int y = PApplet.parseInt(fromY +  (toX - fromX) * t);
			cells.add(new Cell(x, y, cellStatusLazer));
			t += 0.05f;
		}
		return cells;
	}
}
class Level {

	int[][] cells; 

	PImage level;
	PImage imageData;

	int gridWidth;
	int gridHeight;

	PVector scroll;

	int probabilityOfAliveAtStart = 25;

	float timeSpawnStarted = 0;
	float timeSpawnDelay = 10000;

	ArrayList<Cell> lastLazerCells;

	int countMaxBadCells = 20;
	int countMinBadCells = 4;

	boolean spawned = true;

	int timeSpawnDelayAFK = 20000;

	Level() {


		imageData = loadImage("data/corps.png");

		BuildGrid();

		DrawLevel();

		scroll = new PVector(screenX * -0.2f, screenY * -0.55f);

		lastLazerCells = new ArrayList<Cell>();
	}

	public void BuildGrid() {

		imageData.loadPixels();

		gridWidth = PApplet.parseInt(imageData.width/cellSize);
		gridHeight = PApplet.parseInt(imageData.height/cellSize);

  		cells = new int[gridWidth][gridHeight];

  		level = createImage(imageData.width, imageData.height, RGB);
  		level.loadPixels();

	    int pixelCount = imageData.pixels.length;

	    ResetCount();

		for (int p = 0; p < pixelCount; p++) {
			int x = (p % imageData.width) / cellSize;
			int y = floor((p / imageData.width) / cellSize);
			int cellStatus = GetTypeByColor(imageData.pixels[p]);
			cells[x][y] = cellStatus;
			CountCell(cellStatus);
		}

		CheckoutCount();
	}

	public void Update() {
		image(level, scroll.x, scroll.y);
		// DrawPlayer();
		DrawBullets();

		// Spawn new wave of Bad Cells
		if (timeSpawnStarted + timeSpawnDelay < millis()) {
		// || timeSpawnStarted + timeSpawnDelayAFK < millis()) {
			SpawnBad();
			DrawLevel();
			// timeSpawnDelay = 10000 + random(3000);
			// spawned = true;
			timeSpawnStarted = millis();
			waves++;
		}

		// if (spawned && cellCountBad < countMinBadCells) {
		// 	timeSpawnStarted = millis();
		// 	spawned = false;
		// }
	}

	String[] automataObject = {"glider.png", "puffer0.png", "puffer2.png", "spaceship_heavy.png", "spaceship_light.png", "spaceship_middle.png"};

	public void SpawnObject(int x, int y) {
		String randomObject = automataObject[PApplet.parseInt(random(automataObject.length))];
		PImage imageObject = loadImage("data/" + randomObject);
		imageObject.loadPixels();

		int pixelCount = imageObject.pixels.length;
		for (int p = 0; p < pixelCount; p++) {
			int iX = x + p % imageObject.width;
			int iY = y + floor(p / imageObject.width);
			if (IsPositionValid(iX, iY)) {
				cells[iX][iY] = cellStatusBad;
			}
		}

		println("spawn " + randomObject);
	}

	public void SpawnBad() {

		int xx = PApplet.parseInt((random(screenX) + abs(scroll.x)) / cellSize);
		int yy = PApplet.parseInt((random(screenY) + abs(scroll.y)) / cellSize);

		if(random(100) < 50) {

			// Spawn object

			SpawnObject(xx, yy);

		} else {

			// Random square

			int offset = 10 + PApplet.parseInt(random(10));
			int w = xx + offset;
			int h = yy + offset;
			for (int x=xx; x<w; x++) {
				for (int y=yy; y<h; y++) {
					if (IsPositionValid(x, y)) {
						float state = random (100);
						if (state <= probabilityOfAliveAtStart) {
							cells[x][y] = cellStatusBad;
						}
					}
				}
			}
		}
	}

	// void DrawPlayer() {
	//     int pixelCount = player.bodySize * player.bodySize;

	//     PVector playerPos = new PVector(player.position.x, player.position.y);

	//     fill(player.cellColor);

	// 	for (int p = 0; p < pixelCount; p++) {
	// 		int iX = (p % player.bodySize);
	// 		int iY = floor(p/player.bodySize);
	// 		int x = int(playerPos.x + iX * cellSize);
	// 		int y = int(playerPos.y + iY * cellSize);
	// 		color cellColor = player.body[iX][iY].cellColor;
	// 		rect(x, y, cellSize, cellSize);
	// 	}
	// }

	public void DrawBullets() {

		for (int b = 0; b < player.bullets.size(); b++) {

			Bullet bullet = player.bullets.get(b);

			if (bullet.IsAlive()) {

				bullet.position.x += bullet.direction.x * bullet.speed;
				bullet.position.y += bullet.direction.y * bullet.speed;

				if (IsPositionValid(bullet.x, bullet.y))
					cells[bullet.x][bullet.y] = bullet.previousCellStatus;

				int x = PApplet.parseInt(bullet.position.x + abs(scroll.x)) / cellSize;
				int y = PApplet.parseInt(bullet.position.y + abs(scroll.y)) / cellSize;

				if (IsPositionValid(x, y)) {

					bullet.x = x;
					bullet.y = y;
					bullet.previousCellStatus = cells[x][y];

					// Bullet out of Screen
					if (x < 0 || x >= gridWidth || y < 0 || y >= gridHeight ||
						// Out of time
						bullet.timeStarted + bullet.timeDelay < millis()) {

						bullet.Kill();

					// Bullet on screen				
					} else {

						int currentCell = cells[x][y];

						// Check for bad neighbours
						ArrayList<Cell> neighbours = GetNeighbours(x, y, cellStatusBad);

						// Kill bad neighbours
						int badCount = neighbours.size();
						if (badCount > 0) {
							bullet.Kill();
							for (int n = 0; n < badCount; n++) {
								Cell cellNeighbour = neighbours.get(n);
								int nX = PApplet.parseInt(cellNeighbour.x);
								int nY = PApplet.parseInt(cellNeighbour.y);
								cells[nX][nY] = cellStatusDead;
							}
						}
						// Bullet cell /* COMMENT FOR DIRTY MODE */
						else if (currentCell == cellStatusBullet) {
							bullet.Kill();
						}
						// 
						else if (currentCell == cellStatusDead) {

							/* HEAL */

							int neighboursVital = HowManyVitalNeighbours(x, y);
							if (neighboursVital > 0) {
								int averageNeighbour = GetAverageNeighbour(x, y);
								// println("averageNeighbour: "+averageNeighbour);
								bullet.Kill();
								SetNeighbours(x, y, averageNeighbour);
								cells[x][y] = averageNeighbour;
							} else {
								// bullet.AddGhost(x, y, currentCell);
								cells[x][y] = cellStatusBullet;
							}

						} else {

							// bullet.AddGhost(x, y, currentCell);
							cells[x][y] = cellStatusBullet;
						}

					}
				}
			}
		}
		
	}

	public void DrawLevel() {

	    level.loadPixels();

	    int pixelCount = level.pixels.length;

	    ResetCount();

		for (int p = 0; p < pixelCount; p++) {
			int x = (p % imageData.width) / cellSize;
			int y = floor((p/imageData.width) / cellSize);
			level.pixels[p] = GetColorByType(cells[x][y]);
			CountCell(cells[x][y]);
		}
	    level.updatePixels();
	}

	public void CleanLazer() {
		for (int c = 0; c < lastLazerCells.size(); c++) {
			Cell cell = lastLazerCells.get(c);
			cells[cell.x][cell.y] = cell.status;
		}
		lastLazerCells = new ArrayList<Cell>();
	}

	public void FireLazer() {
		CleanLazer();
		int fromX = PApplet.parseInt((player.position.x - scroll.x) / cellSize);
		int fromY = PApplet.parseInt((player.position.y - scroll.y) / cellSize);
		int toX = PApplet.parseInt((mouseX - scroll.x) / cellSize);
		int toY = PApplet.parseInt((mouseY - scroll.y) / cellSize);
		float t = 0.0f;
		while(t < 1.0f) {
			int x = PApplet.parseInt(fromX + (toX - fromX) * t);
			int y = PApplet.parseInt(fromY +  (toY - fromY) * t);
			if (IsPositionValid(x, y) && cells[x][y] != cellStatusBullet) {
				lastLazerCells.add(new Cell(x, y, cells[x][y]));
				cells[x][y] = cellStatusBullet;
			}
			t += 0.05f;
		}
	}

	public boolean IsPositionValid(int xx, int yy) {
		return ((xx>=0)&&(xx<imageData.width/cellSize))&&((yy>=0)&&(yy<imageData.height/cellSize));
	}

	public boolean IsPositionValid(int xx, int yy, int ww, int hh) {
		return ((xx>=0)&&(xx<ww/cellSize))&&((yy>=0)&&(yy<hh/cellSize));
	}

	public int HowManyNeighbours(int x, int y, int cellStatus) { 
		int count = 0;
		for (int xx=x-1; xx<=x+1;xx++)
			for (int yy=y-1; yy<=y+1;yy++)
				if (IsPositionValid(xx, yy)) 
					if (!((xx==x)&&(yy==y)))
						if (cells[xx][yy] == cellStatus)
							count++;
		return count;
	}

	public ArrayList<Cell> GetNeighbours(int x, int y, int cellStatus) { 
		ArrayList<Cell> neighbours = new ArrayList<Cell>();
		for (int xx=x-1; xx<=x+1;xx++)
			for (int yy=y-1; yy<=y+1;yy++)
				if (IsPositionValid(xx, yy)) 
					if (!((xx==x)&&(yy==y)))
						if (cells[xx][yy] == cellStatus)
							neighbours.add(new Cell(xx, yy, cellStatus));
		return neighbours;
	}

	public int HowManyVitalNeighbours(int x, int y) {
		int vitals = 0;
		for (int xx=x-1; xx<=x+1;xx++) {
			for (int yy=y-1; yy<=y+1;yy++) {
				if (IsPositionValid(xx, yy)) {
					if (!((xx==x)&&(yy==y))) {
						int cell = cells[xx][yy];
						if (cell != cellStatusDead && cell != cellStatusBullet && cell != cellStatusBad) {
							vitals++;
						}
					}
				}
			}
		}
		return vitals;
	}

	public int GetAverageNeighbour(int x, int y) { 
		int firstColor = 0;
		IntList neighbours = new IntList();
		for (int xx=x-1; xx<=x+1;xx++) {
			for (int yy=y-1; yy<=y+1;yy++) {
				if (IsPositionValid(xx, yy)) {
					if (!((xx==x)&&(yy==y))) {
						int cell = cells[xx][yy];
						if (cell != cellStatusDead && cell != cellStatusBullet) {
							if (neighbours.hasValue(cell)) {
								return cell;
							} else {
								neighbours.append(cell);
								if (firstColor == 0) firstColor = cell;
							}
						}
					}
				}
			}
		}
		return firstColor;
	}

	public void SetNeighbours(int x, int y, int cellStatus) {
		for (int xx=x-1; xx<=x+1;xx++)
			if (IsPositionValid(xx, y)) 
				if (!(xx==x))
					cells[xx][y] = cellStatus;
		for (int yy=y-1; yy<=y+1;yy++)
			if (IsPositionValid(x, yy)) 
				if (!(yy==y))
					cells[x][yy] = cellStatus;
	}
}
class Player {

	Cell[][] body;
	PVector position;
	float speed = 12;
	int bodySize = 2;
	int cellColor = color(0, 0, 255);

	ArrayList<Bullet> bullets;
	float fireRate = 30;
	float lastFiredTime = 0;
	int bulletRadius = 2;

	Player () {
		position = new PVector(screenX / 2, screenY / 2);
		// body = new Cell[bodySize][bodySize];
		// for (int x = 0; x < bodySize; x++) {
		// 	for (int y = 0; y < bodySize; y++) {
		// 		body[x][y] = new Cell(x, y, color(0, 0, 255));
		// 	}
		// }

		bullets = new ArrayList<Bullet>();
	}

	public void Update() {

		// 
		Draw();

		// MOVE
		Move();

		// BULLETS FIRE
  		if (millis() - lastFiredTime > fireRate && mouseButton == LEFT) {
			Shoot();
			lastFiredTime = millis();
		}
	}

	public void Draw() {
		// noStroke();
		// fill(255);
		// rect(position.x - cellSize * 2, position.y - cellSize, cellSize * 5, cellSize * 3);
		// rect(position.x - cellSize, position.y - cellSize * 2, cellSize * 3, cellSize * 5);
		fill(0xff23fe12);
		rect(position.x - cellSize, position.y - cellSize, cellSize * 3, cellSize * 3);
	}

	float scrollOffset = screenX * 0.4f;

	public void Move() {

		// Moving
		if (keys.get("up") == 1) {
			position.y = max(0, min(position.y - speed, screenY - bodySize * cellSize)); 
		} else if (keys.get("down") == 1) {
			position.y = max(0, min(position.y + speed, screenY - bodySize * cellSize)); 
		}
		if (keys.get("left") == 1) {
			position.x = max(0, min(position.x - speed, screenX - bodySize * cellSize)); 
		} else if (keys.get("right") == 1) {
			position.x = max(0, min(position.x + speed, screenX - bodySize * cellSize)); 
		}

		// Scrolling
		if (position.x < scrollOffset && level.scroll.x < 0) {

			level.scroll.x = min(0, max(level.scroll.x + speed, screenX - level.imageData.width));
			position.x = scrollOffset;

		} else if (position.x > screenX - scrollOffset && level.scroll.x > screenX - level.imageData.width)  {

			level.scroll.x = min(0, max(level.scroll.x - speed, screenX - level.imageData.width));
			position.x = screenX - scrollOffset;

		} 

		// Scrolling
		if (position.y < scrollOffset && level.scroll.y < 0) {

			level.scroll.y = min(0, max(level.scroll.y + speed, screenY - level.imageData.height));
			position.y = scrollOffset;

		} else if (position.y > screenY - scrollOffset && level.scroll.y > screenY - level.imageData.height)  {

			level.scroll.y = min(0, max(level.scroll.y - speed, screenY - level.imageData.height));
			position.y = screenY - scrollOffset;

		} 
		
	}

	public void Shoot() {
		int x = PApplet.parseInt(position.x / cellSize);
		int y = PApplet.parseInt(position.y / cellSize);
		boolean recycling = false;
		for (int b = 0; b < bullets.size(); b++) {
			Bullet bullet = bullets.get(b);
			if (bullet.IsAlive() == false) {
				bullet.Start(position.x, position.y, mouseX - position.x, mouseY - position.y, level.cells[x][y]);
				recycling = true;
				break;
			}
		}
		if (!recycling) {
			bullets.add(new Bullet(position.x, position.y, mouseX - position.x, mouseY - position.y, level.cells[x][y]));
		}
	}

	public int BulletAt(int x, int y) {
		for (int b = 0; b < bullets.size(); b++) {
			Bullet bullet = bullets.get(b);
			float bX = bullet.position.x - level.scroll.x;
			float bY = bullet.position.y - level.scroll.y;
			if (x > PApplet.parseInt(bX - bulletRadius) &&
				x < PApplet.parseInt(bX + bulletRadius) && 
				y > PApplet.parseInt(bY - bulletRadius) && 
				y < PApplet.parseInt(bY + bulletRadius)) {
				return b;
			}
		}
		return -1;
	}

	public void KillBullet(int index) {
		bullets.remove(index);
	}
}


final int cellStatusDead = 0;
final int cellStatusBad = 1;
final int cellStatusHeart = 2;
final int cellStatusVeinRed = 3;
final int cellStatusVeinBlue = 4;
final int cellStatusLung = 5;
final int cellStatusBone = 6;
final int cellStatusSkinLight = 7;
final int cellStatusSkinMedium = 8;
final int cellStatusSkinDark = 9;
final int cellStatusTendon = 10;
final int cellStatusBrain = 11;
final int cellStatusEye = 12;
final int cellStatusSkinShadow = 13;
final int cellStatusOrganYellow = 14;
final int cellStatusInstetinLigh = 15;
final int cellStatusInstetinMedium = 16;
final int cellStatusSkinLeg = 17;
final int cellStatusOrganBrownLight = 18;
final int cellStatusOrganBrownMedium = 19;
final int cellStatusOrganBrownDark = 20;
final int cellStatusOrganBrownVeryLight = 21;
final int cellStatusOrganBrownFade = 22;
final int cellStatusOrganBrownAgain = 23;
final int cellStatusOrganBrownOther = 24;

final int cellStatusBullet = 1000;
final int cellStatusLazer = 9000;

public int GetTypeByColor(int cellColor) {
  int type = 0;
  switch (cellColor){
    case 0:
    case 0xff000000:
      type = cellStatusDead;
      break;
    case 0xffcccccc:
      type = cellStatusBad;
    case 0xffff7c7b:
      type = cellStatusHeart;
      break ;
    case 0xffff057b:
      type = cellStatusVeinRed;
      break;
    case 0xff7777ff:
      type = cellStatusVeinBlue;
      break;
    case 0xffffc57b:
      type = cellStatusLung;
      break;
    case 0xfffff4bd:
      type = cellStatusBone;
      break;
    case 0xff48d24f:
      type = cellStatusBullet;
      break;
    case 0xfff6cbac:
      type = cellStatusSkinLight;
      break;
    case 0xffe2b698:
      type = cellStatusSkinMedium;
      break;
    case 0xffd09170:
      type = cellStatusSkinDark;
      break;
    case 0xfff8d0cb:
      type = cellStatusTendon;
      break;
    case 0xffffc07b:
    case 0xffffbd7b:
    case 0xfffeb97b:
    case 0xfffebf7b:
      type = cellStatusBrain;
      break;
    case 0xffffffff:
      type = cellStatusEye;
      break;
    case 0xffad5632:
      type = cellStatusSkinShadow;
      break;
    case 0xffefd362:
      type = cellStatusOrganYellow;
      break;
    case 0xffeccb86:
      type = cellStatusInstetinLigh;
      break;
    case 0xffd8b572:
      type = cellStatusInstetinMedium;
      break;
    case 0xffecd4bb:
      type = cellStatusSkinLeg;
      break;
    case 0xffc7653d:
      type = cellStatusOrganBrownLight;
      break;
    case 0xffbb5438:
      type = cellStatusOrganBrownMedium;
      break;
    case 0xffa3522e:
      type = cellStatusOrganBrownDark;
      break;
    case 0xffce9773:
      type = cellStatusOrganBrownVeryLight;
      break;
    case 0xffa87b59:
      type = cellStatusOrganBrownFade;
      break;
    case 0xffe7d2c1:
      type = cellStatusOrganBrownAgain;
      break;
    case 0xffb06a49:
      type = cellStatusOrganBrownOther;
      break;
    case 0xff12bf03:
      type = cellStatusLazer;
      break;
  }
  return type;
}

public int GetColorByType(int cellType) {
  int color_ = 0xff000000;
  switch (cellType){
    case cellStatusDead:
      color_ = 0xff000000;
      break;
    case cellStatusBad:
      color_ = 0xffcccccc;
      break;
    case cellStatusHeart:
      color_ = 0xffff7c7b;
      break;
    case cellStatusVeinRed:
      color_ = 0xffff057b;
      break;
    case cellStatusVeinBlue:
      color_ = 0xff7777ff;
      break;
    case cellStatusLung:
      color_ = 0xffffc57b;
      break;
    case cellStatusBone:
      color_ = 0xfffff4bd;
      break;
    case cellStatusBullet:
      color_ = 0xff48d24f;
      break;
    case cellStatusSkinLight:
      color_ = 0xfff6cbac;
      break;
    case cellStatusSkinMedium:
      color_ = 0xffe2b698;
      break;
    case cellStatusSkinDark:
      color_ = 0xffd09170;
      break;
    case cellStatusTendon:
      color_ = 0xfff8d0cb;
      break;
    case cellStatusBrain:
      color_ = 0xffffc07b;
      break;
    case cellStatusEye:
      color_ = 0xffffffff;
      break;
    case cellStatusSkinShadow:
      color_ = 0xffad5632;
      break;
    case cellStatusOrganYellow:
      color_ = 0xffefd362;
      break;
    case cellStatusInstetinLigh:
      color_ = 0xffeccb86;
      break;
    case cellStatusInstetinMedium:
      color_ = 0xffd8b572;
      break;
    case cellStatusSkinLeg:
      color_ = 0xffecd4bb;
      break;
    case cellStatusOrganBrownLight:
      color_ = 0xffc7653d;
      break;
    case cellStatusOrganBrownMedium:
      color_ = 0xffbb5438;
      break;
    case cellStatusOrganBrownDark:
      color_ = 0xffa3522e;
      break;
    case cellStatusOrganBrownVeryLight:
      color_ = 0xffce9773;
      break;
    case cellStatusOrganBrownFade:
      color_ = 0xffa87b59;
      break;
    case cellStatusOrganBrownAgain:
      color_ = 0xffe7d2c1;
      break;
    case cellStatusOrganBrownOther:
      color_ = 0xffb06a49;
      break;
    case cellStatusLazer:
      color_ = 0xff12bf03;
      break;
  }
  return color_;
}

public boolean TestCollision(int x, int y, int xx, int yy, int ww, int hh) {
  return (x > xx && x < xx + ww && y > yy && y < yy + hh);
}

int cellTotalVital = 1;
int cellCountVital = 1;
int cellCountBad = 1;
int cellCountHeart = 1;
int cellCountLung = 1;
int cellCountBrain = 1;
int cellCountSkin = 1;
int cellTotalHeart = 1;
int cellTotalBrain = 1;
int cellTotalSkin = 1;
int cellTotalLung = 1;

public void CountCell(int cellStatus) {
  switch (cellStatus){
    case cellStatusHeart:
      cellCountHeart++;
      break;
    case cellStatusLung:
      cellCountLung++;
      break;
    case cellStatusVeinRed:
      cellCountVital++;
      break;
    case cellStatusBrain:
      cellCountBrain++;
      break;
    case cellStatusSkinLeg:
    case cellStatusSkinShadow:
    case cellStatusSkinDark:
    case cellStatusSkinMedium:
    case cellStatusSkinLight:
      cellCountSkin++;
      break;
    case cellStatusBad:
      cellCountBad++;
      break;
  }
}

public void ResetCount() {
  cellCountBad = 1;
  cellCountVital = 1;
  cellTotalBrain = 1;
  cellTotalSkin = 1;
  cellTotalLung = 1;
  cellTotalHeart = 1;
  // cellCounts = new IntDict();
  // cellTotals = new IntDict();
}

public void CheckoutCount() {
  cellTotalVital = cellCountVital/(cellSize*cellSize);
  cellTotalBrain = cellCountBrain/(cellSize*cellSize);
  cellTotalSkin = cellCountSkin/(cellSize*cellSize);
  cellTotalLung = cellCountLung/(cellSize*cellSize);
  cellTotalHeart = cellCountHeart/(cellSize*cellSize);
  // cellTotals = new IntDict();
  // String[] countKeys = cellCounts.keyArray();
  // int[] countValues = cellCounts.valueArray();
  // for (int c = 0; c < countKeys.length; c++) {
  //   cellTotals.set(countKeys[c], countValues[c]);
  // }
}

public float[] GetRatioCells() {
  float[] cells = new float[5];
  cells[0] = cellCountVital/(cellSize*cellSize) / PApplet.parseFloat(cellTotalVital);
  cells[1] = cellCountBrain/(cellSize*cellSize) / PApplet.parseFloat(cellTotalBrain);
  cells[2] = cellCountSkin/(cellSize*cellSize) / PApplet.parseFloat(cellTotalSkin);
  cells[3] = cellCountLung/(cellSize*cellSize) / PApplet.parseFloat(cellTotalLung);
  cells[4] = cellCountHeart/(cellSize*cellSize) / PApplet.parseFloat(cellTotalHeart);
  return cells;
}

public float GetVitalCellRatio() {
  return cellCountVital/(cellSize*cellSize) / PApplet.parseFloat(cellTotalVital);
  // int countVital = 0;
  // int countTotal = 0;
  // String[] countKeys = cellCounts.keyArray();
  // String[] totalKeys = cellTotals.keyArray();
  // for (int c = 0; c < countKeys.length; c++) 
  //   if (countKeys[c] != "status_0" || countKeys[c] != "status_1") 
  //     countVital++;
  // for (int c = 0; c < totalKeys.length; c++) 
  //   if (totalKeys[c] != "status_0" || totalKeys[c] != "status_1") 
  //     countTotal++;
  // return countVital / float(countTotal);
}

public float GetBadCellRatio() {
  return cellCountBad/(cellSize*cellSize) / PApplet.parseFloat(cellCountVital);
  // int countVital = 0;
  // int countBad = 0;
  // String[] countKeys = cellCounts.keyArray();
  // String[] totalKeys = cellTotals.keyArray();
  // for (int c = 0; c < countKeys.length; c++) {
  //   if (countKeys[c] != "status_0" || countKeys[c] != "status_1") {
  //       countVital++;
  //     } else {
  //       countBad++;
  //     }
  // }
  // return countBad / float(countVital);
}
  static public void main(String[] passedArgs) {
    String[] appletArgs = new String[] { "PixelSurgeon" };
    if (passedArgs != null) {
      PApplet.main(concat(appletArgs, passedArgs));
    } else {
      PApplet.main(appletArgs);
    }
  }
}
